from r2x.api import System
from r2x.config_scenario import Scenario
from r2x.runner import run_exporter, run_parser


def test_round_trip(tmp_folder, reeds_data_folder):
    reeds_infrasys_config = Scenario.from_kwargs(
        name="from_reeds",
        input_model="reeds-US",
        output_model="infrasys",
        run_folder=reeds_data_folder,
        output_folder=tmp_folder,
        solve_year=2050,
        model_year=2050,
        weather_year=2012,
    )
    infrasys_plexos_config = Scenario.from_kwargs(
        name="from_reeds",
        input_model="infrasys",
        output_model="plexos",
        run_folder=tmp_folder,
        output_folder=tmp_folder,
        solve_year=2050,
        model_year=2050,
        weather_year=2012,
    )

    user_dict = {"plexos_category_map": reeds_infrasys_config.input_config.defaults["tech_to_fuel_pm"]}
    plexos_infrasys_config = Scenario.from_kwargs(
        name="from_plexos",
        input_model="plexos",
        output_model="infrasys",
        run_folder=tmp_folder,
        output_folder=tmp_folder,
        solve_year=2050,
        model_year=2050,
        weather_year=2012,
        model_name="model_2012",
        user_dict=user_dict,
    )

    # Export ReEDS to infrasys
    orignal_system_fpath = tmp_folder / f"{reeds_infrasys_config.name}.json"
    original_system, parser = run_parser(reeds_infrasys_config)
    original_system.to_json(orignal_system_fpath)

    # Export infrasys to plexos

    deserialized_original = System.from_json(orignal_system_fpath)
    _ = run_exporter(infrasys_plexos_config, deserialized_original)

    # Export plexos to infrasys
    deserialized_plexos, parser = run_parser(plexos_infrasys_config)

    assert (
        original_system._components.get_num_components()
        == deserialized_plexos._components.get_num_components()
    )

    # Check all components parsed by type are the same
    original_system_components = original_system._component_mgr.get_num_components_by_type()
    deserialized_plexos_components = deserialized_plexos._component_mgr.get_num_components_by_type()
    assert original_system_components == deserialized_plexos_components



"""Plugin to add hurdle rate between regions.

This plugin is only applicable for ReEDs, but could work with similarly arrange data
"""

from argparse import ArgumentParser

from loguru import logger

from r2x.api import System
from r2x.config_scenario import Scenario
from r2x.models.branch import MonitoredLine
from r2x.parser.handler import BaseParser


def cli_arguments(parser: ArgumentParser):
    """CLI arguments for the plugin."""
    parser.add_argument(
        "--hurdle-rate",
        type=float,
        help="Hurdle rate between regions",
    )


def update_system(
    config: Scenario,
    system: System,
    hurdle_rate: float | None = None,
    parser: BaseParser | None = None,
) -> System:
    """Apply hurdle rate between regions.

    This function updates the default hurdle rate for the ReEDS parser using a new file

    Parameters
    ----------
    config : Scenario
        The scenario configuration.
    system : System
        The system object to be updated.
    hurdle_rate : str, optional
        The hurdle rate to apply betwen regions.
    parser : BaseParser, optional
        The parser object used for parsing.
    """
    if not config.output_model == "plexos" or not config.input_model == "reeds-US":
        msg = "Plugin `hurdle_rate.py` is not compatible with a model that is not Plexos."
        raise NotImplementedError(msg)

    if hurdle_rate is None:
        logger.warning("Could not set hurdle rate value. Skipping plugin.")
        return system
    logger.info("Applying hurdle rate to transmission lines")

    if parser is not None:
        assert parser.data.get("hierarchy") is not None, (
            "Did not find hierarchy file on parser. Check parser object."
        )

    for line in system.get_components(MonitoredLine):
        region_to = line.to_bus.load_zone.name
        region_from = line.from_bus.load_zone.name
        if region_to != region_from:
            if previous_hurdle := line.ext.get("Wheeling Charge"):
                logger.debug(
                    "Changing hurdle rate for {} from {} to {}.", line.name, previous_hurdle, hurdle_rate
                )
            line.ext["Wheeling Charge"] = hurdle_rate
            line.ext["Wheeling Charge Back"] = hurdle_rate
    return system



"""Plugin to add annual carbon cap to the model.

This plugin is only applicable for ReEDs, but could work with similarly arrange data
"""

from argparse import ArgumentParser

from loguru import logger

from r2x.api import System
from r2x.config_scenario import Scenario
from r2x.enums import EmissionType
from r2x.models import Emission
from r2x.models.utils import Constraint, ConstraintMap
from r2x.parser.handler import BaseParser
from r2x.units import ureg
from r2x.utils import validate_string


def cli_arguments(parser: ArgumentParser):
    """CLI arguments for the plugin."""
    parser.add_argument(
        "--emission-cap",
        type=float,
        help="Emission cap for the solve year in tonnes. Defaults to CO2 or CO2e.",
    )


def update_system(
    config: Scenario,
    system: System,
    parser: BaseParser | None = None,
    emission_cap: float | None = None,
    default_unit: str = "tonne",
) -> System:
    """Apply an emission cap constraint for the system.

    This function adds to the sytem a :class:`~r2x.models.Constraint`object that is used to set the maximum
    emission per year.

    Parameters
    ----------
    config : Scenario
        The scenario configuration.
    parser : BaseParser
        The parser object used for parsing.
    system : System
        The system object to be updated.
    emission_cap : float or None, optional
        The emission cap value. If None, no cap is applied. Default is None.
    default_unit : str, optional
        The default unit for measurement. Default is 'tonne'.

    Notes
    -----
    When summarizing emissions from either fuels or generators, the metric model
    defines one unit in summary (day, week, month, year) as 1000 of the base units,
    whereas the imperial U.S. model uses 2000 units. Thus, if you define a
    constraint on total emissions over a day, week, month, or year, you must
    enter the limit in the appropriate unit. For example, if the production rate
    is in lb/MWh, then an annual constraint would be in short tons, where one
    short ton equals 2000 lbs. For units in kg/MWh and `emission_cap` in metric tons,
    we multiply by 1000 (`Scalar` property in Plexos).
    """
    if not config.output_model == "plexos":
        msg = "Plugin `emission_cap.py` is not compatible with a model that is not Plexos."
        raise NotImplementedError(msg)

    logger.info("Adding emission cap...")

    if emission_cap is not None:
        logger.debug("Using emission cap value from CLI. Setting emission cap to {}", emission_cap)

    emission_object = EmissionType.CO2  #  This is the default emission object.
    if not any(
        component.emission_type == emission_object
        for component in system.get_supplemental_attributes(Emission)
    ):
        logger.warning("Did not find any emission type to apply emission_cap")
        return system

    if parser is not None and config.input_model == "reeds-US" and emission_cap is None:
        assert parser.data.get("switches") is not None, "Missing switches file from run folder."
        switches = {key: validate_string(value) for key, value in parser.data["switches"].iter_rows()}
        emission_object = EmissionType.CO2E if switches["gsw_annualcapco2e"] else EmissionType.CO2
        assert parser.data.get("co2_cap", None) is not None, "co2_cap not found from ReEDS parser"
        emission_cap = parser.data["co2_cap"]["value"].item()

    if emission_cap is None:
        logger.warning("Could not set emission cap value. Skipping plugin.")
        return system

    emission_cap = ureg.Quantity(emission_cap, default_unit)

    # All of this are Plexos properties that need to be added
    constraint_properties = {
        "Sense": -1,
        "RHS Year": emission_cap,
        "Scalar": 1e3,
        "Penalty Price": 500,
        emission_object: {"Production Coefficient": 1},
    }
    constraint = Constraint(name=f"Annual_{emission_object}_cap", ext=constraint_properties)

    constraint_map = list(system.get_components(ConstraintMap))
    if len(constraint_map) > 1:
        msg = "Multiple constraint maps are not supported yet."
        raise NotImplementedError(msg)

    if not constraint_map:
        constraint_map = ConstraintMap(name="Constraints")
        system.add_component(constraint_map)

    if isinstance(constraint_map, list):
        constraint_map = constraint_map[0]

    constraint_map.mapping[constraint.name].append(emission_object)

    system.add_component(constraint)
    return system



"""Cost related functions."""

from operator import attrgetter
from typing import Annotated

from infrasys.cost_curves import CostCurve, FuelCurve, UnitSystem
from infrasys.models import InfraSysBaseModel
from infrasys.value_curves import LinearCurve
from pydantic import Field, NonNegativeFloat, computed_field


class OperationalCost(InfraSysBaseModel):
    @computed_field  # type: ignore[prop-decorator]
    @property
    def class_type(self) -> str:
        """Create attribute that holds the class name."""
        return type(self).__name__

    @computed_field  # type: ignore[prop-decorator]
    @property
    def variable_type(self) -> str | None:
        """Create attribute that holds the class name."""
        if not getattr(self, "variable", None):
            return None
        return type(getattr(self, "variable")).__name__

    @computed_field  # type: ignore[prop-decorator]
    @property
    def value_curve_type(self) -> str | None:
        """Create attribute that holds the class name."""
        try:
            return type(attrgetter("variable.value_curve")(self)).__name__
        except AttributeError:
            return None

    @computed_field  # type: ignore[prop-decorator]
    @property
    def function_data_type(self) -> str | None:
        """Create attribute that holds the class name."""
        try:
            return type(attrgetter("variable.value_curve.function_data")(self)).__name__
        except AttributeError:
            return None


class RenewableGenerationCost(OperationalCost):
    curtailment_cost: CostCurve | None = None
    variable: CostCurve = CostCurve(value_curve=LinearCurve(0), power_units=UnitSystem.NATURAL_UNITS)


class HydroGenerationCost(OperationalCost):
    fixed: Annotated[
        NonNegativeFloat | None,
        Field(
            description=(
                "Fixed cost of keeping the unit online. "
                "For some cost represenations this field can be duplicative"
            )
        ),
    ] = 0.0
    variable: CostCurve | None = None


class ThermalGenerationCost(OperationalCost):
    """An operational cost for thermal generators.

    It includes fixed cost, variable cost, shut-down cost, and multiple options for start up costs.

    References
    ----------
    .. [1] National Renewable Energy Laboratory. "Thermal Generation Cost Model Library."
       Available: https://nrel-sienna.github.io/PowerSystems.jl/stable/model_library/thermal_generation_cost/
    """

    fixed: Annotated[NonNegativeFloat, Field(description="Cost of using fuel in $ or $/hr.")] = 0.0
    shut_down: Annotated[NonNegativeFloat, Field(description="Cost to turn the unit off")] = 0.0
    start_up: Annotated[NonNegativeFloat, Field(description="Cost to start the unit.")] = 0.0
    variable: Annotated[CostCurve | FuelCurve | None, Field(description="Variable production cost")] = None

    @classmethod
    def example(cls) -> "ThermalGenerationCost":
        return ThermalGenerationCost(
            fixed=0.0,
            shut_down=100.0,
            start_up=100.0,
            variable=FuelCurve(value_curve=LinearCurve(10), power_units=UnitSystem.NATURAL_UNITS),
        )


class StorageCost(OperationalCost):
    charge_variable_cost: CostCurve | None = None
    discharge_variable_cost: CostCurve | None = None
    energy_shortage_cost: Annotated[
        NonNegativeFloat, Field(description="Cost incurred by the model for being short of the energy target")
    ] = 0.0
    energy_surplus_cost: Annotated[NonNegativeFloat, Field(description="Cost of using fuel in $/MWh.")] = 0.0
    fixed: Annotated[NonNegativeFloat, Field(description=" Fixed cost of operating the storage system")] = 0.0
    shut_down: Annotated[NonNegativeFloat | None, Field(description="Cost to turn the unit off")] = 0.0
    start_up: Annotated[NonNegativeFloat | None, Field(description="Cost to start the unit.")] = 0.0



from pathlib import Path
import pytest

from tempfile import NamedTemporaryFile

import polars as pl
from r2x.parser.handler import csv_handler
from r2x.parser.plexos_utils import (
    DATAFILE_COLUMNS,
    get_column_enum,
)


@pytest.mark.parametrize(
    "expected_enum,csv_content",
    [
        (DATAFILE_COLUMNS.NV, "name,value\nTemp,25.5\nLoad,1200\nPressure,101.3"),
        (DATAFILE_COLUMNS.TS_NYV, "name,year,value\nTemp,2030,25.5\nLoad,2030,1200\nPressure,2030,101.3"),
    ],
)
def test_csv_handler(expected_enum, csv_content):
    with NamedTemporaryFile(mode="w+", suffix=".csv", delete=False) as temp_file:
        temp_file.write(csv_content)
        temp_file.seek(0)

        df_csv = csv_handler(Path(temp_file.name))
        assert isinstance(df_csv, pl.DataFrame)
        column_type = get_column_enum(df_csv.columns)
        assert column_type == expected_enum



import pytest
import polars as pl
from datetime import datetime

from r2x.parser.parser_helpers import (
    field_filter,
    fill_missing_timestamps,
    prepare_ext_field,
    reconcile_timeseries,
    resample_data_to_hourly,
)


@pytest.mark.parametrize(
    "property_fields, eligible_fields, expected_valid, expected_extra",
    [
        (
            {"field1": 10, "field2": None, "field3": "hello"},
            {"field1", "field2"},
            {"field1": 10},
            {"field3": "hello"},
        ),
        ({"field1": 10, "field3": "hello"}, {"field2"}, {}, {"field1": 10, "field3": "hello"}),
        ({}, {"field1", "field2"}, {}, {}),
        ({"field1": 10, "field2": 20}, {"field1", "field2"}, {"field1": 10, "field2": 20}, {}),
        ({"field1": None, "field2": 20, "field3": None}, {"field1", "field2"}, {"field2": 20}, {}),
    ],
)
def test_field_filter(property_fields, eligible_fields, expected_valid, expected_extra):
    valid, extra = field_filter(property_fields, eligible_fields)
    assert valid == expected_valid
    assert extra == expected_extra


def test_prepare_ext_field():
    # Test case 1: With extra fields containing eligible and non-eligible types
    valid_fields = {"field1": 10, "field2": "hello"}
    extra_fields = {"field3": [1, 2, 3], "field4": 42, "field5": None}
    result = prepare_ext_field(valid_fields, extra_fields)
    assert result == {"field1": 10, "field2": "hello", "ext": {"field4": 42}}

    # Test case 2: No extra fields
    valid_fields = {"field1": 10, "field2": "hello"}
    extra_fields = {}
    result = prepare_ext_field(valid_fields, extra_fields)
    assert result == {"field1": 10, "field2": "hello", "ext": {}}

    # Test case 3: Various types including eligible and non-eligible
    valid_fields = {"field1": 10, "field2": "hello"}
    extra_fields = {
        "field3": [1, 2, 3],  # Not eligible
        "field4": 42,  # Eligible
        "field5": "world",  # Eligible
        "field6": None,  # Not eligible
        "field7": 3.14,  # Eligible
    }
    result = prepare_ext_field(valid_fields, extra_fields)
    assert result == {
        "field1": 10,
        "field2": "hello",
        "ext": {"field4": 42, "field5": "world", "field7": 3.14},
    }

    # Test case 4: All non-eligible extra fields
    valid_fields = {"field1": 10, "field2": "hello"}
    extra_fields = {
        "field3": [1, 2, 3],  # Not eligible
        "field4": {1: "a"},  # Not eligible
        "field5": None,  # Not eligible
    }
    result = prepare_ext_field(valid_fields, extra_fields)
    assert result == {"field1": 10, "field2": "hello", "ext": {}}


def test_resample_data_to_hourly():
    """Test resampling of half-hourly data to hourly data."""
    # Test case 1: Resampling from half-hourly to hourly
    input_data_1 = pl.DataFrame(
        {
            "year": [2020, 2020],
            "month": [2, 2],
            "day": [28, 28],
            "hour": [0, 0],
            "period": [0, 1],
            "value": [1, 2],
        }
    )
    result_1 = resample_data_to_hourly(input_data_1)
    assert len(result_1) == 1  # Expecting 1 hourly value
    assert result_1["value"].to_list() == [1.5]  # Expected average value

    input_data_2 = pl.DataFrame(
        {
            "year": [2024] * 48,  # Single day, multiple periods
            "month": [10] * 48,
            "day": [24] * 48,
            "period": list(range(1, 49)),  # 1-48 represents 30-minute intervals
            "value": [10] * 48,  # Example values for each period
        }
    )

    result_2 = resample_data_to_hourly(input_data_2)

    # Check the result length and values
    assert len(result_2) == 24  # Expecting 24 hourly values
    assert result_2["value"].to_list() == [10] * 24  # Expected filled values


@pytest.fixture
def hourly_leap_year():
    year = 2020
    return pl.datetime_range(
        datetime(year, 1, 1), datetime(year + 1, 1, 1), interval="1h", eager=True, closed="left"
    ).to_frame("datetime")


@pytest.fixture
def hourly_non_leap_year():
    year = 2021
    return pl.datetime_range(
        datetime(year, 1, 1), datetime(year + 1, 1, 1), interval="1h", eager=True, closed="left"
    ).to_frame("datetime")


def test_fill_missing_timestamps():
    """Test filling missing timestamps and forward filling nulls."""
    year = 2020
    data_file = pl.DataFrame(
        {
            "year": [2020, 2020],
            "month": [1, 1],
            "day": [1, 1],
            "hour": [0, 1],  # Missing hour 1
            "value": [1, 3],
        }
    )

    hourly_time_index = pl.datetime_range(
        datetime(year, 1, 1), datetime(year, 1, 2), interval="1h", eager=True, closed="left"
    ).to_frame("datetime")

    # Call the function
    result = fill_missing_timestamps(data_file, hourly_time_index)

    # Assert that the result contains 24 rows (for each hour of the day)
    assert len(result) == 24

    data_file = pl.DataFrame({"name": ["testname"], "year": [2020], "month": [1], "value": [1]})
    result = fill_missing_timestamps(data_file, hourly_time_index)

    data_file = pl.DataFrame({"year": [2020], "value": [1]})
    with pytest.raises(ValueError):
        _ = fill_missing_timestamps(data_file, hourly_time_index)


def test_reconcile_timeseries_non_leap_year(hourly_non_leap_year, hourly_leap_year):
    # Extract year, month, day, and hour from the datetime column
    data_file = hourly_leap_year.with_columns(
        [
            pl.col("datetime").dt.year().alias("year"),
            pl.col("datetime").dt.month().alias("month"),
            pl.col("datetime").dt.day().alias("day"),
            pl.col("datetime").dt.hour().alias("hour"),
            (pl.arange(0, hourly_leap_year.height)).alias("value"),  # Sequential values
        ]
    )

    # Adjust data
    result = reconcile_timeseries(data_file, hourly_non_leap_year)

    # Expected result should remove Feb 29 data (1416 to 1440)
    assert result.height == 8760
    assert (result["value"] == list(range(1416)) + list(range(1440, 8784))).all()


def test_reconcile_timeseries_leap_year(hourly_non_leap_year, hourly_leap_year):
    # Data file with non-leap year length (8760 hours), leap year hourly_time_index
    data_file = pl.DataFrame(
        {
            "year": [2021] * 8760,
            "month": [2] * 8760,
            "day": [28] * 8760,
            "hour": list(range(8760)),
            "value": list(range(8760)),
        }
    )

    # Leap year hourly_time_index (8784 hours)
    hourly_time_index = pl.datetime_range(
        datetime(2020, 1, 1), datetime(2021, 1, 1), interval="1h", eager=True, closed="left"
    ).to_frame("datetime")

    # Adjust data
    result = reconcile_timeseries(data_file, hourly_time_index)

    # Check that the result has added Feb 29th data
    assert result.height == 8784


def test_reconcile_timeseries_raises_assertion():
    # Empty hourly_time_index
    hourly_time_index = pl.DataFrame()

    # Data file with arbitrary data
    data_file = pl.DataFrame({"year": [2021], "month": [2], "day": [28], "hour": [0], "value": [1]})

    # Check that AssertionError is raised
    with pytest.raises(AssertionError):
        reconcile_timeseries(data_file, hourly_time_index)



import pytest
from pint import Quantity
from r2x.exporter.utils import (
    apply_default_value,
    apply_extract_key,
    apply_flatten_key,
    apply_property_map,
    apply_unnest_key,
    apply_valid_properties,
    apply_pint_deconstruction,
    get_property_magnitude,
)


@pytest.mark.exporter_utils
def test_apply_property_map():
    """Test the apply_property_map function."""
    component = {"voltage": 230, "current": 10}
    property_map = {"voltage": "v", "current": "i"}

    result = apply_property_map(component, property_map)
    assert result == {"v": 230, "i": 10}

    component = {"voltage": 230, "resistance": 50}
    property_map = {"voltage": "v", "current": "i"}
    result = apply_property_map(component, property_map)
    assert result == {"v": 230, "resistance": 50}

    # Test for empty component
    result = apply_property_map({}, property_map)
    assert result == {}


@pytest.mark.exporter_utils
def test_apply_pint_deconstruction():
    """Test the apply_pint_deconstruction function."""
    component = {"length": Quantity(100, "meters"), "time": Quantity(3600, "seconds")}
    unit_map = {"length": "km", "time": "h"}

    result = apply_pint_deconstruction(component, unit_map)
    assert result["length"] == 0.1  # 100 meters to kilometers
    assert result["time"] == 1  # 3600 seconds to hours

    result = apply_pint_deconstruction(component, {})
    assert result["length"] == 100
    assert result["time"] == 3600


@pytest.mark.exporter_utils
def test_apply_valid_properties():
    """Test the apply_valid_properties function."""
    component = {"voltage": 230, "current": 10, "resistance": 50}
    valid_properties = ["voltage", "current"]

    # Test filtering
    result = apply_valid_properties(component, valid_properties)
    assert result == {"voltage": 230, "current": 10}

    component_with_name = {"voltage": 230, "current": 10, "resistance": 50, "name": "Component A"}
    valid_properties_with_name = ["voltage", "current"]

    result_with_name = apply_valid_properties(component_with_name, valid_properties_with_name, add_name=True)
    assert result_with_name == {"voltage": 230, "current": 10, "name": "Component A"}

    result_empty = apply_valid_properties(component, [])
    assert result_empty == {}


@pytest.mark.exporter_utils
def test_get_property_magnitude():
    """Test the get_property_magnitude function."""
    q1 = Quantity(100, "meters")  # Pint Quantity
    q2 = Quantity(50, "kilograms")  # Pint Quantity
    q3 = 200  # Not a Quantity

    assert get_property_magnitude(q1, "kilometers") == 0.1  # Convert 100 meters to kilometers
    assert get_property_magnitude(q2, "grams") == 50000  # Convert 50 kg to grams

    assert get_property_magnitude(q3) == 200  # No conversion for a non-Quantity
    assert get_property_magnitude(q1) == 100  # Magnitude of Quantity without conversion


def test_apply_unnest_key_basic_functionality():
    # Test basic functionality
    component = {"name": "Example", "config": {"type": "A", "value": 10}, "data": {"content": "Some data"}}
    key_map = {"config": "type", "data": "content"}
    result = apply_unnest_key(component, key_map)

    assert result == {"name": "Example", "config": "A", "data": "Some data"}

    # Test no change when key is not in key_map
    component = {"name": "Example", "value": 42}
    key_map = {"config": "type"}
    result = apply_unnest_key(component, key_map)
    assert result == component

    # Test nested dictionary with no corresponding key in key_map
    component = {"name": "Example", "config": {"type": "A", "value": 10}}
    key_map = {"name": "type"}
    result = apply_unnest_key(component, key_map)
    assert result == component


def test_apply_unnest_key_edge_cases():
    # Test missing nested key
    component = {"config": {"value": 10}, "data": {"content": "Some data"}}
    key_map = {"config": "type", "data": "content"}
    result = apply_unnest_key(component, key_map)
    assert result == {"config": {"value": 10}, "data": "Some data"}

    # Test empty input
    assert apply_unnest_key({}, {"config": "type"}) == {}

    # Test empty key_map
    component = {"name": "Example", "config": {"type": "A", "value": 10}}
    assert apply_unnest_key(component, {}) == component


# Parameterized tests
@pytest.mark.parametrize(
    "component,key_map,expected",
    [
        ({"a": {"x": 1, "y": 2}, "b": {"z": 3}}, {"a": "x", "b": "z"}, {"a": 1, "b": 3}),
        (
            {"a": 1, "b": {"x": 2, "y": 3}, "c": "test"},
            {"b": "y", "c": "nonexistent"},
            {"a": 1, "b": 3, "c": "test"},
        ),
        ({"a": {"x": {"nested": "value"}}, "b": 2}, {"a": "x", "b": "y"}, {"a": {"nested": "value"}, "b": 2}),
    ],
)
def parameterized_test(component, key_map, expected):
    assert apply_unnest_key(component, key_map) == expected


def test_flatten_selected_keys():
    d1 = {"x": {"min": 1, "max": 2}, "y": {"min": 5, "max": 10}, "z": 42}
    result1 = apply_flatten_key(d1, {"x"})
    expected1 = {"x_min": 1, "x_max": 2, "y": {"min": 5, "max": 10}, "z": 42}
    assert result1 == expected1

    result2 = apply_flatten_key(d1, {"y"})
    expected2 = {"x": {"min": 1, "max": 2}, "y_min": 5, "y_max": 10, "z": 42}
    assert result2 == expected2

    result3 = apply_flatten_key(d1, set())
    expected3 = d1
    assert result3 == expected3

    result4 = apply_flatten_key(d1, {"x", "y"})
    expected4 = {"x_min": 1, "x_max": 2, "y_min": 5, "y_max": 10, "z": 42}
    assert result4 == expected4


def test_apply_default_value():
    component = {"name": "example", "year": None}
    default_value_map = {"year": 2024, "month": "October"}
    result = apply_default_value(component, default_value_map)
    assert result == {"name": "example", "year": 2024, "month": "October"}

    component = {"name": "example"}
    default_value_map = {"year": 2024, "month": "October"}
    result = apply_default_value(component, default_value_map)
    assert result == {"name": "example", "year": 2024, "month": "October"}

    component = {"name": "example", "year": 2023}
    default_value_map = {"year": 2024, "month": "October"}
    result = apply_default_value(component, default_value_map)
    assert result == {"name": "example", "year": 2023, "month": "October"}

    component = {"name": "example", "year": 2023}
    default_value_map = {}
    result = apply_default_value(component, default_value_map)
    assert result == {"name": "example", "year": 2023}

    component = {}
    default_value_map = {"year": 2024, "month": "October"}
    result = apply_default_value(component, default_value_map)
    assert result == {"year": 2024, "month": "October"}


def test_extract_key():
    component = {"name": "example", "ext": {"TestNested": 1.0}}
    result = apply_extract_key(component, key="ext", keys_to_extract={"TestNested"})
    assert result is not None
    assert result.get("TestNested", None) is not None
    assert result["TestNested"] == 1.0

    component = {"name": "example", "ext": {"TestNested": 1.0, "TestNested2": "test"}}
    result = apply_extract_key(component, key="ext", keys_to_extract={"TestNested", "TestNested2"})
    assert result is not None
    assert result.get("TestNested", None) is not None
    assert result["TestNested"] == 1.0
    assert result.get("TestNested2", None) is not None
    assert result["TestNested2"] == "test"

    component = {"name": "example"}
    result = apply_extract_key(component, key="ext", keys_to_extract={"TestNested", "TestNested2"})
    assert result is not None
    assert result == component

    component = {"name": "example", "ext": {"TestNested": 1.0, "TestNested2": "test"}}
    result = apply_extract_key(component, key="ext", keys_to_extract={"Test"})
    assert result is not None
    assert result == component



import numpy
import pytest
from pint import Quantity

from r2x.api import System
from r2x.config_scenario import Scenario
from r2x.enums import EmissionType
from r2x.models import Emission
from r2x.models.generators import ThermalStandard
from r2x.models.utils import Constraint, ConstraintMap
from r2x.plugins.emission_cap import update_system
from r2x.runner import run_parser, run_plugins


def test_update_system_default(reeds_data_folder, tmp_folder):
    config = Scenario.from_kwargs(
        name="5bus",
        run_folder=reeds_data_folder,
        output_folder=tmp_folder,
        input_model="reeds-US",
        output_model="plexos",
        solve_year=2035,
        weather_year=2012,
    )

    system, parser = run_parser(config)

    new_system = update_system(config=config, parser=parser, system=system)
    assert isinstance(new_system, System)

    constraint = next(iter(new_system.get_components(Constraint)))
    assert constraint
    assert isinstance(constraint, Constraint)
    assert constraint.ext is not None
    assert constraint.ext.get("RHS Year") is not None
    assert isinstance(constraint.ext["RHS Year"], Quantity)
    assert numpy.isclose(constraint.ext["RHS Year"].magnitude, 1.14e9, rtol=1e-02)

    constraint_map = next(iter(new_system.get_components(ConstraintMap)))
    assert constraint_map
    assert isinstance(constraint_map, ConstraintMap)
    assert EmissionType.CO2 in constraint_map.mapping[constraint.name]

    system, parser = run_parser(config)
    new_system = update_system(config=config, parser=parser, system=system, emission_cap=0.0)
    assert isinstance(new_system, System)
    constraint = next(iter(new_system.get_components(Constraint)))
    assert constraint
    assert isinstance(constraint, Constraint)
    assert constraint.ext is not None
    assert constraint.ext.get("RHS Year") is not None
    assert isinstance(constraint.ext["RHS Year"], Quantity)
    assert constraint.ext["RHS Year"].magnitude == 0

    # Test invalid models
    config.output_model = "sienna"
    with pytest.raises(NotImplementedError):
        _ = update_system(config=config, parser=parser, system=system)


def test_no_emission(caplog, infrasys_test_system):
    config = Scenario.from_kwargs(
        name="Pacific",
        input_model="reeds-US",
        output_model="plexos",
        solve_year=2035,
        weather_year=2012,
        emission_cap=0.0,
        plugins=["emission_cap"],
    )

    _ = update_system(config=config, system=infrasys_test_system)
    assert "Did not find any emission" in caplog.text


def test_emission_but_no_cap(caplog):
    system = System(name="Test")
    gen1 = ThermalStandard.example()
    emission = Emission(rate=10, emission_type=EmissionType.CO2)
    system.add_supplemental_attribute(gen1, emission)
    config = Scenario.from_kwargs(
        name="Pacific",
        input_model="reeds-US",
        output_model="plexos",
        solve_year=2035,
        weather_year=2012,
        plugins=["emission_cap"],
    )

    _ = update_system(config=config, system=system)
    assert "Could not set emission cap value" in caplog.text


def test_multiple_constraint_maps():
    system = System(name="Test")
    gen1 = ThermalStandard.example()
    emission = Emission(rate=10, emission_type=EmissionType.CO2)
    constraint_map = ConstraintMap(name="Constraints")
    constraint_map2 = ConstraintMap(name="Constraints")
    system.add_components(constraint_map, constraint_map2)
    system.add_supplemental_attribute(gen1, emission)
    config = Scenario.from_kwargs(
        name="Pacific",
        input_model="reeds-US",
        output_model="plexos",
        solve_year=2035,
        weather_year=2012,
        plugins=["emission_cap"],
    )

    with pytest.raises(NotImplementedError):
        _ = update_system(config=config, system=system, emission_cap=0.0)


def test_update_system_using_cli(reeds_data_folder, tmp_folder):
    config = Scenario.from_kwargs(
        name="Pacific",
        input_model="reeds-US",
        output_model="plexos",
        run_folder=reeds_data_folder,
        output_folder=tmp_folder,
        solve_year=2035,
        weather_year=2012,
        emission_cap=0.0,
        plugins=["emission_cap"],
    )

    system, parser = run_parser(config)
    new_system = run_plugins(config=config, parser=parser, system=system)
    assert isinstance(new_system, System)
    constraint = next(iter(new_system.get_components(Constraint)))
    assert constraint
    assert isinstance(constraint, Constraint)
    assert constraint.ext is not None
    assert constraint.ext.get("RHS Year") is not None
    assert isinstance(constraint.ext["RHS Year"], Quantity)
    assert constraint.ext["RHS Year"].magnitude == 0



"""Electrolyzer representation on PC.

This extension incorporates the load related to the usage of electrolyzer for
each of the ReEDS regions.
"""

# System packages
from datetime import datetime, timedelta

# Third-party packages
import numpy as np
import polars as pl
from infrasys.time_series_models import SingleTimeSeries
from loguru import logger

# Local imports
from r2x.api import System
from r2x.config_scenario import Scenario
from r2x.models import ACBus, Generator, InterruptiblePowerLoad
from r2x.parser.handler import BaseParser
from r2x.parser.polars_helpers import pl_left_multi_join
from r2x.units import ActivePower, FuelPrice

ELECTROLYZER_LOAD_FMAP = "electrolyzer_load"
MONTHLY_H2_FPRICE_FMAP = "h2_fuel_price"


def update_system(parser: BaseParser, config: Scenario, system: System, **_):
    """Modify infrasys system to include electrolyzer load and monthly hydrogen fuel price."""
    logger.info("Adding electrolyzer representation to the system")

    system = electrolyzer_load(config=config, parser=parser, system=system)
    system = hydrogen_fuel_price(config=config, parser=parser, system=system)

    return system


def electrolyzer_load(config: Scenario, parser: BaseParser, system: System) -> System:
    """Add electroylzer load to each region as a fixed load."""
    assert config.input_config.weather_year

    if ELECTROLYZER_LOAD_FMAP not in parser.data:
        logger.warning("No electrolyzer data found on parser. Check parsing filenames.")
        return system

    load_data = parser.data[ELECTROLYZER_LOAD_FMAP]

    if load_data is None:
        logger.warning("No electrolyzer data found on parser. Check parsing filenames.")
        return system

    # Pivot load data to have sum of load for all techs on each column
    load_data_pivot = load_data.pivot(
        index="hour", columns="region", values="load_MW", aggregate_function="sum"
    ).lazy()

    # Create 8760 using hour_map
    hour_map = parser.data["hour_map"]
    total_load_per_region = pl_left_multi_join(hour_map, load_data_pivot).fill_null(0)

    for region in load_data["region"].unique():
        bus = system.get_component(ACBus, name=region)
        max_active_power = ActivePower(0, "MW")
        for component in system.get_components(InterruptiblePowerLoad, filter_func=lambda x: x.bus == bus):
            max_active_power += component.active_power

        # Assert that max active power is greater than 1 MW.
        if max_active_power.magnitude < 1:
            logger.warning("Electrolyzer load for region {} is smaller than 1 MW. Skipping it.", region)
            continue

        # Define new power load and added to the system
        power_load = InterruptiblePowerLoad(name=region, max_active_power=max_active_power, bus=bus)
        system.add_component(power_load)

        ts = SingleTimeSeries.from_array(
            data=ActivePower(total_load_per_region[region], "MW"),
            variable_name="fixed_load",
            initial_time=datetime(year=config.input_config.weather_year, month=1, day=1),
            resolution=timedelta(hours=1),
        )

        user_dict = {"solve_year": config.input_config.weather_year}
        system.add_time_series(ts, power_load, **user_dict)
        logger.debug("Adding electrolyzer load to region: {}", region)
    return system


def hydrogen_fuel_price(config: Scenario, parser: BaseParser, system: System) -> System:
    """Add monthly hydrogen fuel price for generator using hydrogen."""
    if MONTHLY_H2_FPRICE_FMAP not in parser.data:
        logger.warning("No monthly electrolyzer data found on parser. Check parsing filenames.")
        return system

    assert config.input_config.weather_year

    logger.debug("Adding monthly fuel prices for h2 technologies.")
    h2_fprice = parser.data[MONTHLY_H2_FPRICE_FMAP]

    date_time_array = np.arange(
        f"{config.input_config.weather_year}",
        f"{config.input_config.weather_year + 1}",
        dtype="datetime64[h]",
    )[:-24]  # Removing 1 day to match ReEDS convention and converting into a vector

    months = np.array([dt.astype("datetime64[M]").astype(int) % 12 + 1 for dt in date_time_array])

    # Adding fuel price for all hydrogen generators
    for h2_generator in system.get_components(Generator, filter_func=lambda x: "h2" in x.name):
        if h2_generator.bus.name not in h2_fprice["region"]:
            continue
        region_h2_fprice = h2_fprice.filter(pl.col("region") == h2_generator.bus.name)

        month_datetime_series = np.zeros(len(date_time_array), dtype=float)
        for row in region_h2_fprice.iter_rows(named=True):
            month = int(row["month"].strip("m"))
            month_filter = np.where(months == month)
            month_datetime_series[month_filter] = row["h2_price"]
        # Units from monthly hydrogen fuel price are in $/kg
        # Convert $/kG to $/MMbtu ~ 7.5 kg/MMbtu. ~ $7.5/MWh
        month_datetime_series = month_datetime_series * 7.5

        ts = SingleTimeSeries.from_array(
            data=FuelPrice(month_datetime_series, "kg/MWh"),
            variable_name="fuel_price",
            initial_time=datetime(year=config.input_config.weather_year, month=1, day=1),
            resolution=timedelta(hours=1),
        )
        system.add_time_series(ts, h2_generator)

    return system



"""Models related to services."""

from typing import Annotated, Any

from pydantic import Field, NonNegativeFloat, PositiveFloat, field_serializer

from r2x.enums import ReserveDirection, ReserveType
from r2x.models.core import MinMax, Service
from r2x.models.topology import LoadZone
from r2x.units import Percentage


class Reserve(Service):
    """Class representing a reserve contribution."""

    time_frame: Annotated[
        PositiveFloat,
        Field(description="Timeframe in which the reserve is required in seconds"),
    ] = 1e30
    region: (
        Annotated[
            LoadZone,
            Field(description="LoadZone where reserve requirement is required."),
        ]
        | None
    ) = None
    vors: Annotated[
        float,
        Field(description="Value of reserve shortage in $/MW. Any positive value as as soft constraint."),
    ] = -1
    duration: (
        Annotated[
            PositiveFloat,
            Field(description="Time over which the required response must be maintained in seconds."),
        ]
        | None
    ) = None
    reserve_type: ReserveType
    load_risk: (
        Annotated[
            Percentage,
            Field(
                description="Proportion of Load that contributes to the requirement.",
            ),
        ]
        | None
    ) = None
    # ramp_rate: float | None = None  # NOTE: Maybe we do not need this.
    max_requirement: float = 0  # Should we specify which variable is the time series for?
    direction: ReserveDirection

    @classmethod
    def example(cls) -> "Reserve":
        return Reserve(
            name="ExampleReserve",
            region=LoadZone.example(),
            direction=ReserveDirection.UP,
            reserve_type=ReserveType.REGULATION,
        )


class VariableReserve(Reserve):
    time_frame: Annotated[
        NonNegativeFloat,
        Field(description="Timeframe in which the reserve is required in seconds"),
    ] = 0.0
    requirement: Annotated[
        NonNegativeFloat | None,
        Field(
            description="the value of required reserves in p.u (SYSTEM_BASE), validation range: (0, nothing)"
        ),
    ]
    sustained_time: Annotated[
        NonNegativeFloat,
        Field(description="the time in seconds reserve contribution must sustained at a specified level"),
    ] = 3600.0
    max_output_fraction: Annotated[
        NonNegativeFloat,
        Field(
            ge=0.0,
            le=1.0,
            description="the time in seconds reserve contribution must sustained at a specified level",
        ),
    ] = 1.0
    max_participation_factor: Annotated[
        NonNegativeFloat,
        Field(
            ge=0.0,
            le=1.0,
            description="the maximum portion [0, 1.0] of the reserve that can be contributed per device",
        ),
    ] = 1.0
    deployed_fraction: Annotated[
        NonNegativeFloat,
        Field(
            ge=0.0,
            le=1.0,
            description="Fraction of service procurement that is assumed to be actually deployed.",
        ),
    ] = 3600.0


class TransmissionInterface(Service):
    """Component representing a collection of branches that make up an interface or corridor.

    It can be specified between different :class:`Area` or :class:`LoadZone`.
    The interface can be used to constrain the power flow across it
    """

    active_power_flow_limits: Annotated[
        MinMax, Field(description="Minimum and maximum active power flow limits on the interface (MW)")
    ]
    direction_mapping: Annotated[
        dict[str, int],
        Field(
            description=(
                "Dictionary of the line names in the interface and their direction of flow (1 or -1) "
                "relative to the flow of the interface"
            )
        ),
    ]

    @classmethod
    def example(cls) -> "TransmissionInterface":
        return TransmissionInterface(
            name="ExampleTransmissionInterface",
            active_power_flow_limits=MinMax(-100, 100),
            direction_mapping={"line-01": 1, "line-02": -2},
        )

    @field_serializer("active_power_flow_limits")
    def serialize_active_power_limits(self, min_max: MinMax) -> dict[str, Any]:
        if min_max is not None:
            return {"min": min_max.min, "max": min_max.max}
